@@ -113,26 +113,6 @@ CK_RV ProcClose(void)
     return CKR_OK;
 }
 
-CK_RV APILock(API_Slot_t *sltp)
-{
-    if (pthread_mutex_lock(&sltp->api_mutex)) {
-        TRACE_ERROR("Lock failed.\n");
-        return CKR_CANT_LOCK;
-    }
-
-    return CKR_OK;
-}
-
-CK_RV APIUnLock(API_Slot_t *sltp)
-{
-    if (pthread_mutex_unlock(&sltp->api_mutex)) {
-        TRACE_ERROR("Unlock failed.\n");
-        return CKR_CANT_LOCK;
-    }
-
-    return CKR_OK;
-}
-
 unsigned long AddToSessionList(ST_SESSION_T *pSess)
 {
     unsigned long handle;
@@ -144,9 +124,14 @@ unsigned long AddToSessionList(ST_SESSION_T *pSess)
 
 void RemoveFromSessionList(CK_SESSION_HANDLE handle)
 {
-    bt_node_free(&(Anchor->sess_btree), handle, free);
+    bt_node_free(&(Anchor->sess_btree), handle, TRUE);
 }
 
+struct closeme_arg {
+    CK_SLOT_ID slot_id;
+    CK_BBOOL in_fork_initializer;
+};
+
 /* CloseMe
  *
  * Callback function used to close an individual session for a slot
@@ -155,23 +140,24 @@ void CloseMe(STDLL_TokData_t *tokdata, void *node_value,
              unsigned long node_handle, void *arg)
 {
     CK_RV rv;
-    CK_SLOT_ID slot_id = *(CK_SLOT_ID *) arg;
+    struct closeme_arg *closeme_arg = (struct closeme_arg *) arg;
     ST_SESSION_T *s = (ST_SESSION_T *) node_value;
     API_Slot_t *sltp;
     STDLL_FcnList_t *fcn;
 
     UNUSED(tokdata);
 
-    if (s->slotID == slot_id) {
+    if (s->slotID == closeme_arg->slot_id) {
         /* the single ugliest part about moving to a binary tree: these are the
          * guts of the C_CloseSession function, copied here without tests for
          * validity, since if we're here, they must already have been valid */
-        sltp = &(Anchor->SltList[slot_id]);
+        sltp = &(Anchor->SltList[closeme_arg->slot_id]);
         fcn = sltp->FcnList;
-        rv = fcn->ST_CloseSession(sltp->TokData, s);
+        rv = fcn->ST_CloseSession(sltp->TokData, s,
+                                  closeme_arg->in_fork_initializer);
         if (rv == CKR_OK) {
-            decr_sess_counts(slot_id);
-            bt_node_free(&(Anchor->sess_btree), node_handle, free);
+            decr_sess_counts(closeme_arg->slot_id, s->rw_session);
+            bt_node_free(&(Anchor->sess_btree), node_handle, TRUE);
         }
     }
 }
@@ -183,40 +169,39 @@ void CloseMe(STDLL_TokData_t *tokdata, void *node_value,
  * Once all the nodes are closed, we check to see if the tree is empty and if
  * so, destroy it
  */
-CK_RV CloseAllSessions(CK_SLOT_ID slot_id)
+void CloseAllSessions(CK_SLOT_ID slot_id, CK_BBOOL in_fork_initializer)
 {
     API_Slot_t *sltp = &(Anchor->SltList[slot_id]);
+    struct closeme_arg arg;
 
-    if (APILock(sltp) != CKR_OK)
-        return CKR_CANT_LOCK;
+    arg.slot_id = slot_id;
+    arg.in_fork_initializer = in_fork_initializer;
 
     /* for every node in the API-level session tree, call CloseMe on it */
     bt_for_each_node(sltp->TokData, &(Anchor->sess_btree), CloseMe,
-                     (void *) &slot_id);
+                     (void *)&arg);
 
-    if (bt_is_empty(&(Anchor->sess_btree)))
-        bt_destroy(&(Anchor->sess_btree), NULL);
-
-    if (APIUnLock(sltp) != CKR_OK)
-        return CKR_CANT_LOCK;
-
-    return CKR_OK;
 }
 
 int Valid_Session(CK_SESSION_HANDLE handle, ST_SESSION_T *rSession)
 {
     ST_SESSION_T *tmp;
+    int rc;
 
     tmp = bt_get_node_value(&(Anchor->sess_btree), handle);
     if (tmp) {
         rSession->slotID = tmp->slotID;
         rSession->sessionh = tmp->sessionh;
+        rSession->rw_session = tmp->rw_session;
     }
+    rc = tmp ? TRUE : FALSE;
+    bt_put_node_value(&(Anchor->sess_btree), tmp);
+    tmp = NULL;
 
-    return (tmp ? TRUE : FALSE);
+    return rc;
 }
 
-int API_Initialized()
+int API_Initialized(void)
 {
     if (Anchor == NULL)
         return FALSE;
@@ -241,17 +226,18 @@ int slot_present(CK_SLOT_ID id)
     return TRUE;
 }
 
-void get_sess_count(CK_SLOT_ID slotID, CK_ULONG *ret)
+void get_sess_counts(CK_SLOT_ID slotID, CK_ULONG *ret, CK_ULONG *rw_ret)
 {
     Slot_Mgr_Shr_t *shm;
 
     shm = Anchor->SharedMemP;
     ProcLock();
     *ret = shm->slot_global_sessions[slotID];
+    *rw_ret = shm->slot_global_rw_sessions[slotID];
     ProcUnLock();
 }
 
-void incr_sess_counts(CK_SLOT_ID slotID)
+void incr_sess_counts(CK_SLOT_ID slotID, CK_BBOOL rw_session)
 {
     Slot_Mgr_Shr_t *shm;
 #ifdef PKCS64
@@ -266,14 +252,18 @@ void incr_sess_counts(CK_SLOT_ID slotID)
     ProcLock();
 
     shm->slot_global_sessions[slotID]++;
+    if (rw_session)
+        shm->slot_global_rw_sessions[slotID]++;
 
     procp = &shm->proc_table[Anchor->MgrProcIndex];
     procp->slot_session_count[slotID]++;
+    if (rw_session)
+        procp->slot_rw_session_count[slotID]++;
 
     ProcUnLock();
 }
 
-void decr_sess_counts(CK_SLOT_ID slotID)
+void decr_sess_counts(CK_SLOT_ID slotID, CK_BBOOL rw_session)
 {
     Slot_Mgr_Shr_t *shm;
 #ifdef PKCS64
@@ -290,10 +280,16 @@ void decr_sess_counts(CK_SLOT_ID slotID)
     if (shm->slot_global_sessions[slotID] > 0) {
         shm->slot_global_sessions[slotID]--;
     }
+    if (rw_session && shm->slot_global_rw_sessions[slotID] > 0) {
+        shm->slot_global_rw_sessions[slotID]--;
+    }
 
     procp = &shm->proc_table[Anchor->MgrProcIndex];
     if (procp->slot_session_count[slotID] > 0) {
-        procp->slot_session_count[slotID]++;
+        procp->slot_session_count[slotID]--;
+    }
+    if (rw_session && procp->slot_rw_session_count[slotID] > 0) {
+        procp->slot_rw_session_count[slotID]--;
     }
 
     ProcUnLock();
@@ -321,39 +317,11 @@ int sessions_exist(CK_SLOT_ID slotID)
     return numSessions != 0;
 }
 
-// Terminates all sessions associated with a given process
-// this cleans up any lingering sessions with the process
-// and does not
-//
-// It is only called from the C_Finalize routine
-void Terminate_All_Process_Sessions()
-{
-    CK_SLOT_ID id;
-    CK_RV rv;
-
-    TRACE_DEBUG("Terminate_All_Process_Sessions\n");
-    for (id = 0; id < NUMBER_SLOTS_MANAGED; id++) {
-        // Check if the slot is present in the slot manager
-        // if not just skip it...
-        if (slot_present(id) == TRUE) {
-            rv = C_CloseAllSessions(id);
-        } else {
-            continue;
-        }
-        // If the return code is not OK, we are really hosed
-        // since we are terminating the session.
-        // For now we will just log it
-        if (rv != CKR_OK) {
-            TRACE_DEBUG("Terminate_All_Process_Sessions RV %lx\n", rv);
-        }
-    }
-}
-
 // Register the process with PKCSSLOTD in the shared memory.
 // This call must be made with the API Global Mutex Locked
 // and the Anchor control block initialized with the
 // shared memory.  No checking for shared memory validity is done
-int API_Register()
+int API_Register(void)
 {
     long int reuse = -1, free = -1;
     Slot_Mgr_Shr_t *shm;
@@ -385,12 +353,12 @@ int API_Register()
             // before the slot manager garbage collection can performed.
             // To eliminate the race condition between garbage collection
             // the lock should protect us.
-            // This should be a VERY rare (if ever) occurance, given the
+            // This should be a VERY rare (if ever) occurrence, given the
             // way AIX deals with re-allocation of PID;s, however if this
             // ever gets ported over to another platform we want to deal
             // with this accordingly since it may re-use pids differently
             // (Linux appears to re-use pids more rapidly).
-            if (procp->proc_id == getpid()) {
+            if (procp->proc_id == Anchor->ClientCred.real_pid) {
                 if (reuse == -1) {
                     reuse = indx;
                 }
@@ -409,7 +377,7 @@ int API_Register()
         return FALSE;
     }
     // check if we are reusing a control block or taking the first free.
-    // Since th mutex is helt, we don;t have to worry about some other
+    // Since the mutex is held, we don;t have to worry about some other
     // process grabbing the slot...  Garbage collection from
     // the slotd should not affect this since it will grab the mutex
     // before doing its thing.
@@ -427,13 +395,13 @@ int API_Register()
     memset((char *) procp, 0, sizeof(Slot_Mgr_Proc_t));
 #endif
     procp->inuse = TRUE;
-    procp->proc_id = getpid();
+    procp->proc_id = Anchor->ClientCred.real_pid;
     procp->reg_time = time(NULL);
 
     Anchor->MgrProcIndex = indx;
 
-    TRACE_DEVEL("API_Register MgrProcIndc %d  pid %ld \n", procp->proc_id,
-                (long int) Anchor->MgrProcIndex);
+    TRACE_DEVEL("API_Register MgrProcIndc %ld (real) pid %d \n",
+                (long int) Anchor->MgrProcIndex, procp->proc_id);
 
     //??? What to do about the Mutex and cond variable
     //Does initializing them in the slotd allow for them to not be
@@ -448,7 +416,7 @@ int API_Register()
 // This call must be made with the API Global Mutex Locked
 // and the Anchor control block initialized with the
 // shared memory.  No checking for shared memory validity is done
-void API_UnRegister()
+void API_UnRegister(void)
 {
     Slot_Mgr_Shr_t *shm;
 
@@ -484,7 +452,7 @@ void API_UnRegister()
     ProcUnLock();
 }
 
-void DL_UnLoad(API_Slot_t *sltp, CK_SLOT_ID slotID)
+void DL_UnLoad(API_Slot_t *sltp, CK_SLOT_ID slotID, CK_BBOOL inchildforkinit)
 {
     Slot_Mgr_Socket_t *shData = &(Anchor->SocketDataP);
 #ifdef PKCS64
@@ -493,6 +461,15 @@ void DL_UnLoad(API_Slot_t *sltp, CK_SLOT_ID slotID)
     Slot_Info_t *sinfp;
 #endif
 
+    if (sltp->TokData) {
+#ifdef ENABLE_LOCKS
+        pthread_rwlock_destroy(&sltp->TokData->sess_list_rwlock);
+#endif
+        pthread_mutex_destroy(&sltp->TokData->login_mutex);
+        free(sltp->TokData);
+        sltp->TokData = NULL;
+    }
+
     sinfp = &(shData->slot_info[slotID]);
 
     if (sinfp->present == FALSE) {
@@ -501,6 +478,8 @@ void DL_UnLoad(API_Slot_t *sltp, CK_SLOT_ID slotID)
     if (!sltp->dlop_p) {
         return;
     }
+    if (inchildforkinit)
+        return;
     // Call the routine to properly unload the DLL
     DL_Unload(sltp);
 
@@ -584,11 +563,10 @@ void DL_Unload(API_Slot_t *sltp)
     sltp->dlop_p = NULL;
     sltp->pSTfini = NULL;
     sltp->pSTcloseall = NULL;
-
-    pthread_mutex_destroy(&sltp->api_mutex);
 }
 
-int DL_Load_and_Init(API_Slot_t *sltp, CK_SLOT_ID slotID)
+int DL_Load_and_Init(API_Slot_t *sltp, CK_SLOT_ID slotID, policy_t policy,
+                     statistics_t statistics)
 {
     Slot_Mgr_Socket_t *shData = &(Anchor->SocketDataP);
 #ifdef PKCS64
@@ -596,13 +574,11 @@ int DL_Load_and_Init(API_Slot_t *sltp, CK_SLOT_ID slotID)
 #else
     Slot_Info_t *sinfp;
 #endif
-
-    int (*pSTinit) ();
-    void (*pSTfini) ();
+    CK_RV (*pSTinit)(API_Slot_t *, CK_SLOT_ID, SLOT_INFO *,
+                    struct trace_handle_t);
     CK_RV rv;
-    int dll_len, dl_index;
+    int dl_index;
     DLL_Load_t *dllload;
-    pthread_mutexattr_t attr;
 
     // Get pointer to shared memory from the anchor block
     //
@@ -614,7 +590,34 @@ int DL_Load_and_Init(API_Slot_t *sltp, CK_SLOT_ID slotID)
         return FALSE;
     }
 
-    if ((dll_len = strlen(sinfp->dll_location))) {
+    if (sltp->TokData != NULL) {
+        TRACE_ERROR("Already initialized.\n");
+        return FALSE;
+    }
+
+    /*
+     * Create separate memory area for each token specific data
+     */
+    sltp->TokData = (STDLL_TokData_t *) calloc(1, sizeof(STDLL_TokData_t));
+    if (!sltp->TokData) {
+        TRACE_ERROR("Allocating host memory failed.\n");
+        return FALSE;
+    }
+    sltp->TokData->slot_id = slotID;
+    sltp->TokData->real_pid = Anchor->ClientCred.real_pid;
+    sltp->TokData->real_uid = Anchor->ClientCred.real_uid;
+    sltp->TokData->real_gid = Anchor->ClientCred.real_gid;
+    sltp->TokData->ro_session_count = 0;
+    sltp->TokData->global_login_state = CKS_RO_PUBLIC_SESSION;
+#ifdef ENABLE_LOCKS
+    pthread_rwlock_init(&sltp->TokData->sess_list_rwlock, NULL);
+#endif
+    pthread_mutex_init(&sltp->TokData->login_mutex, NULL);
+    sltp->TokData->policy = policy;
+    sltp->TokData->mechtable_funcs = &mechtable_funcs;
+    sltp->TokData->statistics = statistics;
+    
+    if (strlen(sinfp->dll_location) > 0) {
         // Check if this DLL has been loaded already.. If so, just increment
         // the counter in the dllload structure and copy the data to
         // the slot pointer.
@@ -632,49 +635,31 @@ int DL_Load_and_Init(API_Slot_t *sltp, CK_SLOT_ID slotID)
     }
 
     if (!sltp->dlop_p) {
-        TRACE_DEBUG("DL_Load_and_Init pointer %p\n", sltp->dlop_p);
-
+        TRACE_DEBUG("DL_Load_and_Init pointer NULL\n");
+        DL_UnLoad(sltp, slotID, FALSE);
         return FALSE;
     }
 
     *(void **)(&pSTinit) = dlsym(sltp->dlop_p, "ST_Initialize");
     if (!pSTinit) {
         // Unload the DLL
-        DL_Unload(sltp);
+        DL_UnLoad(sltp, slotID, FALSE);
         return FALSE;
     }
-
-    if (pthread_mutexattr_init(&attr)) {
-        TRACE_ERROR("Mutex attribute init failed.\n");
-        DL_Unload(sltp);
-        return FALSE;
-    }
-    if (pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE)) {
-        TRACE_ERROR("Mutex attribute set failed.\n");
-        DL_Unload(sltp);
-        return FALSE;
-    }
-    if (pthread_mutex_init(&sltp->api_mutex, &attr)) {
-        TRACE_ERROR("Mutex init failed.\n");
-        DL_Unload(sltp);
-        return FALSE;
-    }
-
     // Returns true or false
     rv = pSTinit(sltp, slotID, sinfp, trace);
     TRACE_DEBUG("return from STDDLL Init = %lx\n", rv);
 
     if (rv != CKR_OK) {
         // clean up and unload
-        DL_Unload(sltp);
+        DL_UnLoad(sltp, slotID, FALSE);
         sltp->DLLoaded = FALSE;
         return FALSE;
     } else {
         sltp->DLLoaded = TRUE;
+        sinfp->pk_slot.flags |= CKF_TOKEN_PRESENT;
         // Check if a SC_Finalize function has been exported
-        *(void **)(&pSTfini) = dlsym(sltp->dlop_p, "SC_Finalize");
-        sltp->pSTfini = pSTfini;
-
+        *(void **)(&sltp->pSTfini) = dlsym(sltp->dlop_p, "SC_Finalize");
         *(void **)(&sltp->pSTcloseall) =
             dlsym(sltp->dlop_p, "SC_CloseAllSessions");
         return TRUE;
