@@ -1288,6 +1288,7 @@ CK_RV ckm_des_key_gen(STDLL_TokData_t *tokdata, TEMPLATE *tmpl)
     else
         memcpy(value_attr->pValue, des_key, DES_KEY_SIZE);
     free(des_key);
+    des_key = NULL;
 
     key_type_attr->type = CKA_KEY_TYPE;
     key_type_attr->ulValueLen = sizeof(CK_KEY_TYPE);
@@ -1346,106 +1347,6 @@ err:
     return rc;
 }
 
-#if !(NOCDMF)
-
-//
-//
-CK_RV ckm_cdmf_key_gen(STDLL_TokData_t *tokdata, TEMPLATE *tmpl)
-{
-    CK_ATTRIBUTE *value_attr = NULL;
-    CK_ATTRIBUTE *key_type_attr = NULL;
-    CK_ATTRIBUTE *class_attr = NULL;
-    CK_ATTRIBUTE *local_attr = NULL;
-    CK_BYTE cdmf_key[DES_KEY_SIZE];
-    CK_ULONG repl_len, expected_repl_len;
-    CK_ULONG rc;
-
-    repl_len = expected_repl_len = DES_KEY_SIZE;
-
-    rc = communicate(PK_CDMF_KEYGEN,
-                     NULL, 0, cdmf_key, &repl_len, NULL, 0, NULL, 0);
-
-    if (rc == CKR_OK) {
-        if (repl_len != expected_repl_len)
-            return CKR_GENERAL_ERROR;
-    }
-
-    value_attr = (CK_ATTRIBUTE *) malloc(sizeof(CK_ATTRIBUTE) + DES_KEY_SIZE);
-    key_type_attr =
-        (CK_ATTRIBUTE *) malloc(sizeof(CK_ATTRIBUTE) + sizeof(CK_KEY_TYPE));
-    class_attr =
-        (CK_ATTRIBUTE *) malloc(sizeof(CK_ATTRIBUTE) + sizeof(CK_OBJECT_CLASS));
-    local_attr =
-        (CK_ATTRIBUTE *) malloc(sizeof(CK_ATTRIBUTE) + sizeof(CK_BBOOL));
-
-    if (!value_attr || !key_type_attr || !class_attr || !local_attr) {
-        TRACE_ERROR("%s\n", ock_err(ERR_HOST_MEMORY));
-        rc = CKR_HOST_MEMORY;
-        goto error;
-    }
-
-    value_attr->type = CKA_VALUE;
-    value_attr->ulValueLen = DES_KEY_SIZE;
-    value_attr->pValue = (CK_BYTE *) value_attr + sizeof(CK_ATTRIBUTE);
-    memcpy(value_attr->pValue, cdmf_key, DES_KEY_SIZE);
-
-    key_type_attr->type = CKA_KEY_TYPE;
-    key_type_attr->ulValueLen = sizeof(CK_KEY_TYPE);
-    key_type_attr->pValue = (CK_BYTE *) key_type_attr + sizeof(CK_ATTRIBUTE);
-    *(CK_KEY_TYPE *) key_type_attr->pValue = CKK_CDMF;
-
-    class_attr->type = CKA_CLASS;
-    class_attr->ulValueLen = sizeof(CK_OBJECT_CLASS);
-    class_attr->pValue = (CK_BYTE *) class_attr + sizeof(CK_ATTRIBUTE);
-    *(CK_OBJECT_CLASS *) class_attr->pValue = CKO_SECRET_KEY;
-
-    local_attr->type = CKA_LOCAL;
-    local_attr->ulValueLen = sizeof(CK_BBOOL);
-    local_attr->pValue = (CK_BYTE *) local_attr + sizeof(CK_ATTRIBUTE);
-    *(CK_BBOOL *) local_attr->pValue = TRUE;
-
-    rc = template_update_attribute(tmpl, value_attr);
-    if (rc != CKR_OK) {
-        TRACE_ERROR("template_update_attribute failed\n");
-        goto error;
-    }
-    value_attr = NULL;
-    rc = template_update_attribute(tmpl, key_type_attr);
-    if (rc != CKR_OK) {
-        TRACE_ERROR("template_update_attribute failed\n");
-        goto error;
-    }
-    key_type_attr = NULL;
-    rc = template_update_attribute(tmpl, class_attr);
-    if (rc != CKR_OK) {
-        TRACE_ERROR("template_update_attribute failed\n");
-        goto error;
-    }
-    class_attr = NULL;
-    rc = template_update_attribute(tmpl, local_attr);
-    if (rc != CKR_OK) {
-        TRACE_ERROR("template_update_attribute failed\n");
-        goto error;
-    }
-    local_attr = NULL;
-
-    return CKR_OK;
-
-error:
-    if (value_attr)
-        free(value_attr);
-    if (key_type_attr)
-        free(key_type_attr);
-    if (class_attr)
-        free(class_attr);
-    if (local_attr)
-        free(local_attr);
-
-    return rc;
-}
-
-#endif
-
 //
 //
 CK_RV ckm_des_ecb_encrypt(STDLL_TokData_t *tokdata,
