@@ -35,22 +35,21 @@ CK_RV rsa_get_key_info(OBJECT *key_obj, CK_ULONG *mod_bytes,
     CK_RV rc;
     CK_ATTRIBUTE *attr;
 
-    rc = template_attribute_find(key_obj->template, CKA_MODULUS, &attr);
-    if (rc == FALSE) {
+    rc = template_attribute_get_non_empty(key_obj->template, CKA_MODULUS,
+                                          &attr);
+    if (rc != CKR_OK) {
         TRACE_ERROR("Could not find CKA_MODULUS in the template\n");
-        return CKR_FUNCTION_FAILED;
+        return rc;
     }
 
     *mod_bytes = attr->ulValueLen;
 
-    rc = template_attribute_find(key_obj->template, CKA_CLASS, &attr);
-    if (rc == FALSE) {
+    rc = template_attribute_get_ulong(key_obj->template, CKA_CLASS, keyclass);
+    if (rc != CKR_OK) {
         TRACE_ERROR("Could not find CKA_CLASS in the template\n");
-        return CKR_FUNCTION_FAILED;
+        return rc;
     }
 
-    *keyclass = *(CK_OBJECT_CLASS *) attr->pValue;
-
     return CKR_OK;
 }
 
@@ -178,6 +177,7 @@ CK_RV rsa_parse_block(CK_BYTE *in_data,
 {
     CK_ULONG i;
     CK_RV rc = CKR_OK;
+    int found = 0;
 
     if (!in_data || !out_data || !out_data_len) {
         TRACE_ERROR("%s received bad argument(s)\n", __func__);
@@ -230,9 +230,13 @@ CK_RV rsa_parse_block(CK_BYTE *in_data,
      */
     case 0:
         for (i = 2; i <= (in_data_len - 2); i++) {
-            if (in_data[i] != (CK_BYTE) 0)
+            if (in_data[i] != (CK_BYTE) 0) {
+                found = 1;
                 break;
+            }
         }
+        if (!found)
+            i++;
         break;
     /*
      * For block type 01, they shall have value FF.
@@ -243,6 +247,7 @@ CK_RV rsa_parse_block(CK_BYTE *in_data,
             if (in_data[i] != (CK_BYTE) 0xff) {
                 if (in_data[i] == (CK_BYTE) 0) {
                     i++;
+                    found = 1;
                     break;
                 }
 
@@ -250,6 +255,8 @@ CK_RV rsa_parse_block(CK_BYTE *in_data,
                 return CKR_ENCRYPTED_DATA_INVALID;
             }
         }
+        if (!found)
+            i++;
         break;
     /*
      * For block type 02, they shall be pseudorandomly generated and
@@ -261,9 +268,12 @@ CK_RV rsa_parse_block(CK_BYTE *in_data,
         for (i = 2; i <= (in_data_len - 2); i++) {
             if (in_data[i] == (CK_BYTE) 0) {
                 i++;
+                found = 1;
                 break;
             }
         }
+        if (!found)
+            i++;
         break;
     default:
         TRACE_ERROR("%s\n", ock_err(ERR_ENCRYPTED_DATA_INVALID));
@@ -281,7 +291,7 @@ CK_RV rsa_parse_block(CK_BYTE *in_data,
         return CKR_ENCRYPTED_DATA_INVALID;
     }
 
-    if (in_data_len <= i) {
+    if (in_data_len < 2) {
         TRACE_ERROR("%s\n", ock_err(ERR_ENCRYPTED_DATA_INVALID));
         return CKR_ENCRYPTED_DATA_INVALID;
     }
@@ -297,32 +307,6 @@ CK_RV rsa_parse_block(CK_BYTE *in_data,
     return rc;
 }
 
-/* helper function for rsa-oaep */
-CK_RV get_mgf_mech(CK_RSA_PKCS_MGF_TYPE mgf, CK_MECHANISM_TYPE *mech)
-{
-    switch (mgf) {
-    case CKG_MGF1_SHA1:
-        *mech = CKM_SHA_1;
-        break;
-    case CKG_MGF1_SHA224:
-        *mech = CKM_SHA224;
-        break;
-	case CKG_MGF1_SHA256:
-        *mech = CKM_SHA256;
-        break;
-    case CKG_MGF1_SHA384:
-        *mech = CKM_SHA384;
-        break;
-    case CKG_MGF1_SHA512:
-        *mech = CKM_SHA512;
-        break;
-    default:
-        return CKR_MECHANISM_INVALID;
-    }
-
-    return CKR_OK;
-}
-
 //
 //
 CK_RV rsa_pkcs_encrypt(STDLL_TokData_t *tokdata,
@@ -342,7 +326,7 @@ CK_RV rsa_pkcs_encrypt(STDLL_TokData_t *tokdata,
 
     rc = object_mgr_find_in_map1(tokdata, ctx->key, &key_obj, READ_LOCK);
     if (rc != CKR_OK) {
-        TRACE_ERROR("Failed to acquire key from specified handle");
+        TRACE_ERROR("Failed to acquire key from specified handle.\n");
         if (rc == CKR_OBJECT_HANDLE_INVALID)
             return CKR_KEY_HANDLE_INVALID;
         else
@@ -419,7 +403,7 @@ CK_RV rsa_pkcs_decrypt(STDLL_TokData_t *tokdata,
 
     rc = object_mgr_find_in_map1(tokdata, ctx->key, &key_obj, READ_LOCK);
     if (rc != CKR_OK) {
-        TRACE_ERROR("Failed to acquire key from specified handle");
+        TRACE_ERROR("Failed to acquire key from specified handle.\n");
         if (rc == CKR_OBJECT_HANDLE_INVALID)
             return CKR_KEY_HANDLE_INVALID;
         else
@@ -504,7 +488,7 @@ CK_RV rsa_oaep_crypt(STDLL_TokData_t *tokdata, SESSION *sess,
 
     rc = object_mgr_find_in_map1(tokdata, ctx->key, &key_obj, READ_LOCK);
     if (rc != CKR_OK) {
-        TRACE_ERROR("Failed to acquire key from specified handle");
+        TRACE_ERROR("Failed to acquire key from specified handle.\n");
         if (rc == CKR_OBJECT_HANDLE_INVALID)
             return CKR_KEY_HANDLE_INVALID;
         else
@@ -592,6 +576,10 @@ CK_RV rsa_oaep_crypt(STDLL_TokData_t *tokdata, SESSION *sess,
             goto done;
         }
 
+        /* Release obj lock, token specific rsa-oaep may re-acquire the lock */
+        object_put(tokdata, key_obj, TRUE);
+        key_obj = NULL;
+
         rc = token_specific.t_rsa_oaep_encrypt(tokdata, ctx, in_data,
                                                in_data_len, out_data,
                                                out_data_len, hash, hlen);
@@ -615,6 +603,10 @@ CK_RV rsa_oaep_crypt(STDLL_TokData_t *tokdata, SESSION *sess,
             goto done;
         }
 
+        /* Release obj lock, token specific rsa-oaep may re-acquire the lock */
+        object_put(tokdata, key_obj, TRUE);
+        key_obj = NULL;
+
         rc = token_specific.t_rsa_oaep_decrypt(tokdata, ctx, in_data,
                                                in_data_len, out_data,
                                                out_data_len, hash, hlen);
@@ -650,7 +642,7 @@ CK_RV rsa_pkcs_sign(STDLL_TokData_t *tokdata,
     }
     rc = object_mgr_find_in_map1(tokdata, ctx->key, &key_obj, READ_LOCK);
     if (rc != CKR_OK) {
-        TRACE_ERROR("Failed to acquire key from specified handle");
+        TRACE_ERROR("Failed to acquire key from specified handle.\n");
         if (rc == CKR_OBJECT_HANDLE_INVALID)
             return CKR_KEY_HANDLE_INVALID;
         else
@@ -727,7 +719,7 @@ CK_RV rsa_pkcs_verify(STDLL_TokData_t *tokdata,
 
     rc = object_mgr_find_in_map1(tokdata, ctx->key, &key_obj, READ_LOCK);
     if (rc != CKR_OK) {
-        TRACE_ERROR("Failed to acquire key from specified handle");
+        TRACE_ERROR("Failed to acquire key from specified handle.\n");
         if (rc == CKR_OBJECT_HANDLE_INVALID)
             return CKR_KEY_HANDLE_INVALID;
         else
@@ -796,7 +788,7 @@ CK_RV rsa_pkcs_verify_recover(STDLL_TokData_t *tokdata,
     }
     rc = object_mgr_find_in_map1(tokdata, ctx->key, &key_obj, READ_LOCK);
     if (rc != CKR_OK) {
-        TRACE_ERROR("Failed to acquire key from specified handle");
+        TRACE_ERROR("Failed to acquire key from specified handle.\n");
         if (rc == CKR_OBJECT_HANDLE_INVALID)
             return CKR_KEY_HANDLE_INVALID;
         else
@@ -867,7 +859,7 @@ CK_RV rsa_x509_encrypt(STDLL_TokData_t *tokdata,
 
     rc = object_mgr_find_in_map1(tokdata, ctx->key, &key_obj, READ_LOCK);
     if (rc != CKR_OK) {
-        TRACE_ERROR("Failed to acquire key from specified handle");
+        TRACE_ERROR("Failed to acquire key from specified handle.\n");
         if (rc == CKR_OBJECT_HANDLE_INVALID)
             return CKR_KEY_HANDLE_INVALID;
         else
@@ -944,7 +936,7 @@ CK_RV rsa_x509_decrypt(STDLL_TokData_t *tokdata,
 
     rc = object_mgr_find_in_map1(tokdata, ctx->key, &key_obj, READ_LOCK);
     if (rc != CKR_OK) {
-        TRACE_ERROR("Failed to acquire key from specified handle");
+        TRACE_ERROR("Failed to acquire key from specified handle.\n");
         if (rc == CKR_OBJECT_HANDLE_INVALID)
             return CKR_KEY_HANDLE_INVALID;
         else
@@ -1036,7 +1028,7 @@ CK_RV rsa_x509_sign(STDLL_TokData_t *tokdata,
     }
     rc = object_mgr_find_in_map1(tokdata, ctx->key, &key_obj, READ_LOCK);
     if (rc != CKR_OK) {
-        TRACE_ERROR("Failed to acquire key from specified handle");
+        TRACE_ERROR("Failed to acquire key from specified handle.\n");
         if (rc == CKR_OBJECT_HANDLE_INVALID)
             return CKR_KEY_HANDLE_INVALID;
         else
@@ -1112,7 +1104,7 @@ CK_RV rsa_x509_verify(STDLL_TokData_t *tokdata,
 
     rc = object_mgr_find_in_map1(tokdata, ctx->key, &key_obj, READ_LOCK);
     if (rc != CKR_OK) {
-        TRACE_ERROR("Failed to acquire key from specified handle");
+        TRACE_ERROR("Failed to acquire key from specified handle.\n");
         if (rc == CKR_OBJECT_HANDLE_INVALID)
             return CKR_KEY_HANDLE_INVALID;
         else
@@ -1182,7 +1174,7 @@ CK_RV rsa_x509_verify_recover(STDLL_TokData_t *tokdata,
     }
     rc = object_mgr_find_in_map1(tokdata, ctx->key, &key_obj, READ_LOCK);
     if (rc != CKR_OK) {
-        TRACE_ERROR("Failed to acquire key from specified handle");
+        TRACE_ERROR("Failed to acquire key from specified handle.\n");
         if (rc == CKR_OBJECT_HANDLE_INVALID)
             return CKR_KEY_HANDLE_INVALID;
         else
@@ -1261,7 +1253,7 @@ CK_RV rsa_pss_sign(STDLL_TokData_t *tokdata, SESSION *sess,
 
     rc = object_mgr_find_in_map1(tokdata, ctx->key, &key_obj, READ_LOCK);
     if (rc != CKR_OK) {
-        TRACE_ERROR("Failed to acquire key from specified handle");
+        TRACE_ERROR("Failed to acquire key from specified handle.\n");
         if (rc == CKR_OBJECT_HANDLE_INVALID)
             return CKR_KEY_HANDLE_INVALID;
         else
@@ -1321,6 +1313,10 @@ CK_RV rsa_pss_sign(STDLL_TokData_t *tokdata, SESSION *sess,
         goto done;
     }
 
+    /* Release obj lock, token specific rsa_pss may re-acquire the lock */
+    object_put(tokdata, key_obj, TRUE);
+    key_obj = NULL;
+
     rc = token_specific.t_rsa_pss_sign(tokdata, sess, ctx, in_data, in_data_len,
                                        out_data, out_data_len);
     if (rc != CKR_OK)
@@ -1345,7 +1341,7 @@ CK_RV rsa_pss_verify(STDLL_TokData_t *tokdata, SESSION *sess,
 
     rc = object_mgr_find_in_map1(tokdata, ctx->key, &key_obj, READ_LOCK);
     if (rc != CKR_OK) {
-        TRACE_ERROR("Failed to acquire key from specified handle");
+        TRACE_ERROR("Failed to acquire key from specified handle.\n");
         if (rc == CKR_OBJECT_HANDLE_INVALID)
             return CKR_KEY_HANDLE_INVALID;
         else
@@ -1379,6 +1375,10 @@ CK_RV rsa_pss_verify(STDLL_TokData_t *tokdata, SESSION *sess,
         goto done;
     }
 
+    /* Release obj lock, token specific rsa_pss may re-acquire the lock */
+    object_put(tokdata, key_obj, TRUE);
+    key_obj = NULL;
+
     rc = token_specific.t_rsa_pss_verify(tokdata, sess, ctx, in_data,
                                          in_data_len, signature, sig_len);
     if (rc != CKR_OK)
@@ -1440,16 +1440,17 @@ CK_RV rsa_hash_pss_sign(STDLL_TokData_t *tokdata, SESSION *sess,
         return CKR_MECHANISM_PARAM_INVALID;
     }
 
-    rc = digest_mgr_init(tokdata, sess, &digest_ctx, &digest_mech);
+    rc = digest_mgr_init(tokdata, sess, &digest_ctx, &digest_mech, FALSE);
     if (rc != CKR_OK) {
         TRACE_DEVEL("Digest Mgr Init failed.\n");
         return rc;
     }
 
-    rc = digest_mgr_digest(tokdata, sess, length_only, &digest_ctx,
+    rc = digest_mgr_digest(tokdata, sess, FALSE, &digest_ctx,
                            in_data, in_data_len, hash, &hlen);
     if (rc != CKR_OK) {
         TRACE_DEVEL("Digest Mgr Digest failed.\n");
+        digest_mgr_cleanup(tokdata, sess, &digest_ctx);
         return rc;
     }
 
@@ -1458,7 +1459,8 @@ CK_RV rsa_hash_pss_sign(STDLL_TokData_t *tokdata, SESSION *sess,
     sign_mech.ulParameterLen = ctx->mech.ulParameterLen;
     sign_mech.pParameter = ctx->mech.pParameter;
 
-    rc = sign_mgr_init(tokdata, sess, &sign_ctx, &sign_mech, FALSE, ctx->key);
+    rc = sign_mgr_init(tokdata, sess, &sign_ctx, &sign_mech, FALSE, ctx->key,
+                       FALSE);
     if (rc != CKR_OK) {
         TRACE_DEVEL("Sign Mgr Init failed.\n");
         goto done;
@@ -1470,7 +1472,7 @@ CK_RV rsa_hash_pss_sign(STDLL_TokData_t *tokdata, SESSION *sess,
         TRACE_DEVEL("Sign Mgr Sign failed.\n");
 
 done:
-    sign_mgr_cleanup(&sign_ctx);
+    sign_mgr_cleanup(tokdata, sess, &sign_ctx);
 
     return rc;
 }
@@ -1514,11 +1516,12 @@ CK_RV rsa_hash_pss_update(STDLL_TokData_t *tokdata, SESSION *sess,
         digest_mech.ulParameterLen = 0;
         digest_mech.pParameter = NULL;
 
-        rc = digest_mgr_init(tokdata, sess, digest_ctx, &digest_mech);
+        rc = digest_mgr_init(tokdata, sess, digest_ctx, &digest_mech, FALSE);
         if (rc != CKR_OK) {
             TRACE_DEVEL("Digest Mgr Init failed.\n");
             return rc;
         }
+        ctx->state_unsaveable |= digest_ctx->state_unsaveable;
     }
 
     rc = digest_mgr_digest_update(tokdata, sess, digest_ctx, in_data,
@@ -1549,6 +1552,13 @@ CK_RV rsa_hash_pss_sign_final(STDLL_TokData_t *tokdata, SESSION *sess,
 
     digest_ctx = (DIGEST_CONTEXT *) ctx->context;
 
+    if (digest_ctx->active == FALSE) {
+        rc = rsa_hash_pss_update(tokdata, sess, ctx, NULL, 0);
+        TRACE_DEVEL("rsa_hash_pss_update\n");
+        if (rc != 0)
+            return rc;
+    }
+
     rc = get_sha_size(digest_ctx->mech.mechanism, &hlen);
     if (rc != CKR_OK) {
         TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
@@ -1567,7 +1577,8 @@ CK_RV rsa_hash_pss_sign_final(STDLL_TokData_t *tokdata, SESSION *sess,
     sign_mech.ulParameterLen = ctx->mech.ulParameterLen;
     sign_mech.pParameter = ctx->mech.pParameter;
 
-    rc = sign_mgr_init(tokdata, sess, &sign_ctx, &sign_mech, FALSE, ctx->key);
+    rc = sign_mgr_init(tokdata, sess, &sign_ctx, &sign_mech, FALSE, ctx->key,
+                       FALSE);
     if (rc != CKR_OK) {
         TRACE_DEVEL("Sign Mgr Init failed.\n");
         goto done;
@@ -1579,7 +1590,7 @@ CK_RV rsa_hash_pss_sign_final(STDLL_TokData_t *tokdata, SESSION *sess,
         TRACE_DEVEL("Sign Mgr Sign failed.\n");
 
 done:
-    sign_mgr_cleanup(&sign_ctx);
+    sign_mgr_cleanup(tokdata, sess, &sign_ctx);
 
     return rc;
 }
@@ -1632,7 +1643,7 @@ CK_RV rsa_hash_pss_verify(STDLL_TokData_t *tokdata, SESSION *sess,
         return CKR_MECHANISM_PARAM_INVALID;
     }
 
-    rc = digest_mgr_init(tokdata, sess, &digest_ctx, &digest_mech);
+    rc = digest_mgr_init(tokdata, sess, &digest_ctx, &digest_mech, FALSE);
     if (rc != CKR_OK) {
         TRACE_DEVEL("Digest Mgr Init failed.\n");
         return rc;
@@ -1642,6 +1653,7 @@ CK_RV rsa_hash_pss_verify(STDLL_TokData_t *tokdata, SESSION *sess,
                            in_data_len, hash, &hlen);
     if (rc != CKR_OK) {
         TRACE_DEVEL("Digest Mgr Digest failed.\n");
+        digest_mgr_cleanup(tokdata, sess, &digest_ctx);
         return rc;
     }
 
@@ -1651,7 +1663,7 @@ CK_RV rsa_hash_pss_verify(STDLL_TokData_t *tokdata, SESSION *sess,
     verify_mech.pParameter = ctx->mech.pParameter;
 
     rc = verify_mgr_init(tokdata, sess, &verify_ctx, &verify_mech, FALSE,
-                         ctx->key);
+                         ctx->key, FALSE);
     if (rc != CKR_OK) {
         TRACE_DEVEL("Verify Mgr Init failed.\n");
         goto done;
@@ -1663,7 +1675,7 @@ CK_RV rsa_hash_pss_verify(STDLL_TokData_t *tokdata, SESSION *sess,
         TRACE_DEVEL("Verify Mgr Verify failed.\n");
 
 done:
-    verify_mgr_cleanup(&verify_ctx);
+    verify_mgr_cleanup(tokdata, sess, &verify_ctx);
 
     return rc;
 }
@@ -1688,6 +1700,13 @@ CK_RV rsa_hash_pss_verify_final(STDLL_TokData_t *tokdata, SESSION *sess,
 
     digest_ctx = (DIGEST_CONTEXT *) ctx->context;
 
+    if (digest_ctx->active == FALSE) {
+        rc = rsa_hash_pss_update(tokdata, sess, ctx, NULL, 0);
+        TRACE_DEVEL("rsa_hash_pss_update\n");
+        if (rc != 0)
+            return rc;
+    }
+
     rc = get_sha_size(digest_ctx->mech.mechanism, &hlen);
     if (rc != CKR_OK) {
         TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
@@ -1706,7 +1725,7 @@ CK_RV rsa_hash_pss_verify_final(STDLL_TokData_t *tokdata, SESSION *sess,
     verify_mech.pParameter = ctx->mech.pParameter;
 
     rc = verify_mgr_init(tokdata, sess, &verify_ctx, &verify_mech, FALSE,
-                         ctx->key);
+                         ctx->key, FALSE);
     if (rc != CKR_OK) {
         TRACE_DEVEL("Verify Mgr Init failed.\n");
         goto done;
@@ -1718,7 +1737,7 @@ CK_RV rsa_hash_pss_verify_final(STDLL_TokData_t *tokdata, SESSION *sess,
         TRACE_DEVEL("Verify Mgr Verify failed.\n");
 
 done:
-    verify_mgr_cleanup(&verify_ctx);
+    verify_mgr_cleanup(tokdata, sess, &verify_ctx);
 
     return rc;
 }
@@ -1756,12 +1775,14 @@ CK_RV rsa_hash_pkcs_sign(STDLL_TokData_t *tokdata,
     memset(&digest_ctx, 0x0, sizeof(digest_ctx));
     memset(&sign_ctx, 0x0, sizeof(sign_ctx));
 
+#if !(NOMD2)
     if (ctx->mech.mechanism == CKM_MD2_RSA_PKCS) {
         digest_mech.mechanism = CKM_MD2;
         oid = ber_AlgMd2;
         oid_len = ber_AlgMd2Len;
-
-    } else if (ctx->mech.mechanism == CKM_MD5_RSA_PKCS) {
+    } else
+#endif
+    if (ctx->mech.mechanism == CKM_MD5_RSA_PKCS) {
         digest_mech.mechanism = CKM_MD5;
         oid = ber_AlgMd5;
         oid_len = ber_AlgMd5Len;
@@ -1790,16 +1811,17 @@ CK_RV rsa_hash_pkcs_sign(STDLL_TokData_t *tokdata,
     digest_mech.ulParameterLen = 0;
     digest_mech.pParameter = NULL;
 
-    rc = digest_mgr_init(tokdata, sess, &digest_ctx, &digest_mech);
+    rc = digest_mgr_init(tokdata, sess, &digest_ctx, &digest_mech, FALSE);
     if (rc != CKR_OK) {
         TRACE_DEVEL("Digest Mgr Init failed.\n");
         return rc;
     }
     hash_len = sizeof(hash);
-    rc = digest_mgr_digest(tokdata, sess, length_only, &digest_ctx, in_data,
+    rc = digest_mgr_digest(tokdata, sess, FALSE, &digest_ctx, in_data,
                            in_data_len, hash, &hash_len);
     if (rc != CKR_OK) {
         TRACE_DEVEL("Digest Mgr Digest failed.\n");
+        digest_mgr_cleanup(tokdata, sess, &digest_ctx);
         return rc;
     }
     // build the BER-encodings
@@ -1827,7 +1849,8 @@ CK_RV rsa_hash_pkcs_sign(STDLL_TokData_t *tokdata,
     sign_mech.ulParameterLen = 0;
     sign_mech.pParameter = NULL;
 
-    rc = sign_mgr_init(tokdata, sess, &sign_ctx, &sign_mech, FALSE, ctx->key);
+    rc = sign_mgr_init(tokdata, sess, &sign_ctx, &sign_mech, FALSE, ctx->key,
+                       FALSE);
     if (rc != CKR_OK) {
         TRACE_DEVEL("Sign Mgr Init failed.\n");
         goto error;
@@ -1842,7 +1865,7 @@ error:
         free(octet_str);
     if (ber_data)
         free(ber_data);
-    sign_mgr_cleanup(&sign_ctx);
+    sign_mgr_cleanup(tokdata, sess, &sign_ctx);
 
     return rc;
 }
@@ -1866,9 +1889,12 @@ CK_RV rsa_hash_pkcs_sign_update(STDLL_TokData_t *tokdata,
     context = (RSA_DIGEST_CONTEXT *) ctx->context;
 
     if (context->flag == FALSE) {
+#if !(NOMD2)
         if (ctx->mech.mechanism == CKM_MD2_RSA_PKCS)
             digest_mech.mechanism = CKM_MD2;
-        else if (ctx->mech.mechanism == CKM_MD5_RSA_PKCS)
+        else
+#endif
+        if (ctx->mech.mechanism == CKM_MD5_RSA_PKCS)
             digest_mech.mechanism = CKM_MD5;
         else if (ctx->mech.mechanism == CKM_SHA224_RSA_PKCS)
             digest_mech.mechanism = CKM_SHA224;
@@ -1885,12 +1911,13 @@ CK_RV rsa_hash_pkcs_sign_update(STDLL_TokData_t *tokdata,
         digest_mech.pParameter = NULL;
 
         rc = digest_mgr_init(tokdata, sess, &context->hash_context,
-                             &digest_mech);
+                             &digest_mech, FALSE);
         if (rc != CKR_OK) {
             TRACE_DEVEL("Digest Mgr Init failed.\n");
             return rc;
         }
         context->flag = TRUE;
+        ctx->state_unsaveable |= context->hash_context.state_unsaveable;
     }
 
     rc = digest_mgr_digest_update(tokdata, sess, &context->hash_context,
@@ -1934,11 +1961,14 @@ CK_RV rsa_hash_pkcs_verify(STDLL_TokData_t *tokdata,
     memset(&digest_ctx, 0x0, sizeof(digest_ctx));
     memset(&verify_ctx, 0x0, sizeof(verify_ctx));
 
+#if !(NOMD2)
     if (ctx->mech.mechanism == CKM_MD2_RSA_PKCS) {
         digest_mech.mechanism = CKM_MD2;
         oid = ber_AlgMd2;
         oid_len = ber_AlgMd2Len;
-    } else if (ctx->mech.mechanism == CKM_MD5_RSA_PKCS) {
+    } else
+#endif
+    if (ctx->mech.mechanism == CKM_MD5_RSA_PKCS) {
         digest_mech.mechanism = CKM_MD5;
         oid = ber_AlgMd5;
         oid_len = ber_AlgMd5Len;
@@ -1968,7 +1998,7 @@ CK_RV rsa_hash_pkcs_verify(STDLL_TokData_t *tokdata,
     digest_mech.ulParameterLen = 0;
     digest_mech.pParameter = NULL;
 
-    rc = digest_mgr_init(tokdata, sess, &digest_ctx, &digest_mech);
+    rc = digest_mgr_init(tokdata, sess, &digest_ctx, &digest_mech, FALSE);
     if (rc != CKR_OK) {
         TRACE_DEVEL("Digest Mgr Init failed.\n");
         return rc;
@@ -1978,6 +2008,7 @@ CK_RV rsa_hash_pkcs_verify(STDLL_TokData_t *tokdata,
                            in_data_len, hash, &hash_len);
     if (rc != CKR_OK) {
         TRACE_DEVEL("Digest Mgr Digest failed.\n");
+        digest_mgr_cleanup(tokdata, sess, &digest_ctx);
         return rc;
     }
     // Build the BER encoding
@@ -2005,7 +2036,7 @@ CK_RV rsa_hash_pkcs_verify(STDLL_TokData_t *tokdata,
     verify_mech.pParameter = NULL;
 
     rc = verify_mgr_init(tokdata, sess, &verify_ctx, &verify_mech, FALSE,
-                         ctx->key);
+                         ctx->key, FALSE);
     if (rc != CKR_OK) {
         TRACE_DEVEL("Verify Mgr Init failed.\n");
         goto done;
@@ -2019,7 +2050,7 @@ done:
         free(octet_str);
     if (ber_data)
         free(ber_data);
-    sign_mgr_cleanup(&verify_ctx);
+    sign_mgr_cleanup(tokdata, sess, &verify_ctx);
 
     return rc;
 }
@@ -2042,9 +2073,12 @@ CK_RV rsa_hash_pkcs_verify_update(STDLL_TokData_t *tokdata,
     context = (RSA_DIGEST_CONTEXT *) ctx->context;
 
     if (context->flag == FALSE) {
+#if !(NOMD2)
         if (ctx->mech.mechanism == CKM_MD2_RSA_PKCS)
             digest_mech.mechanism = CKM_MD2;
-        else if (ctx->mech.mechanism == CKM_MD5_RSA_PKCS)
+        else
+#endif
+        if (ctx->mech.mechanism == CKM_MD5_RSA_PKCS)
             digest_mech.mechanism = CKM_MD5;
         else if (ctx->mech.mechanism == CKM_SHA224_RSA_PKCS)
             digest_mech.mechanism = CKM_SHA224;
@@ -2061,12 +2095,13 @@ CK_RV rsa_hash_pkcs_verify_update(STDLL_TokData_t *tokdata,
         digest_mech.pParameter = NULL;
 
         rc = digest_mgr_init(tokdata, sess, &context->hash_context,
-                             &digest_mech);
+                             &digest_mech, FALSE);
         if (rc != CKR_OK) {
             TRACE_DEVEL("Digest Mgr Init failed.\n");
             return rc;
         }
         context->flag = TRUE;
+        ctx->state_unsaveable |= context->hash_context.state_unsaveable;
     }
 
     rc = digest_mgr_digest_update(tokdata, sess, &context->hash_context,
@@ -2107,10 +2142,13 @@ CK_RV rsa_hash_pkcs_sign_final(STDLL_TokData_t *tokdata,
         return CKR_FUNCTION_FAILED;
     }
 
+#if !(NOMD2)
     if (ctx->mech.mechanism == CKM_MD2_RSA_PKCS) {
         oid = ber_AlgMd2;
         oid_len = ber_AlgMd2Len;
-    } else if (ctx->mech.mechanism == CKM_MD5_RSA_PKCS) {
+    } else
+#endif
+    if (ctx->mech.mechanism == CKM_MD5_RSA_PKCS) {
         oid = ber_AlgMd5;
         oid_len = ber_AlgMd5Len;
     } else if (ctx->mech.mechanism == CKM_SHA224_RSA_PKCS) {
@@ -2134,6 +2172,13 @@ CK_RV rsa_hash_pkcs_sign_final(STDLL_TokData_t *tokdata,
 
     context = (RSA_DIGEST_CONTEXT *) ctx->context;
 
+    if (context->flag == FALSE) {
+        rc = rsa_hash_pkcs_sign_update(tokdata, sess, ctx, NULL, 0);
+        TRACE_DEVEL("rsa_hash_pkcs_sign_update\n");
+        if (rc != 0)
+            return rc;
+    }
+
     hash_len = sizeof(hash);
     rc = digest_mgr_digest_final(tokdata, sess, length_only,
                                  &context->hash_context, hash, &hash_len);
@@ -2166,7 +2211,8 @@ CK_RV rsa_hash_pkcs_sign_final(STDLL_TokData_t *tokdata,
     sign_mech.ulParameterLen = 0;
     sign_mech.pParameter = NULL;
 
-    rc = sign_mgr_init(tokdata, sess, &sign_ctx, &sign_mech, FALSE, ctx->key);
+    rc = sign_mgr_init(tokdata, sess, &sign_ctx, &sign_mech, FALSE, ctx->key,
+                       FALSE);
     if (rc != CKR_OK) {
         TRACE_DEVEL("Sign Mgr Init failed.\n");
         goto done;
@@ -2185,7 +2231,7 @@ done:
         free(octet_str);
     if (ber_data)
         free(ber_data);
-    sign_mgr_cleanup(&sign_ctx);
+    sign_mgr_cleanup(tokdata, sess, &sign_ctx);
 
     return rc;
 }
@@ -2215,10 +2261,13 @@ CK_RV rsa_hash_pkcs_verify_final(STDLL_TokData_t *tokdata,
         TRACE_ERROR("%s received bad argument(s)\n", __func__);
         return CKR_FUNCTION_FAILED;
     }
+#if !(NOMD2)
     if (ctx->mech.mechanism == CKM_MD2_RSA_PKCS) {
         oid = ber_AlgMd2;
         oid_len = ber_AlgMd2Len;
-    } else if (ctx->mech.mechanism == CKM_MD5_RSA_PKCS) {
+    } else
+#endif
+    if (ctx->mech.mechanism == CKM_MD5_RSA_PKCS) {
         oid = ber_AlgMd5;
         oid_len = ber_AlgMd5Len;
     } else if (ctx->mech.mechanism == CKM_SHA224_RSA_PKCS) {
@@ -2242,6 +2291,13 @@ CK_RV rsa_hash_pkcs_verify_final(STDLL_TokData_t *tokdata,
 
     context = (RSA_DIGEST_CONTEXT *) ctx->context;
 
+    if (context->flag == FALSE) {
+        rc = rsa_hash_pkcs_verify_update(tokdata, sess, ctx, NULL, 0);
+        TRACE_DEVEL("rsa_hash_pkcs_verify_update\n");
+        if (rc != 0)
+            return rc;
+    }
+
     hash_len = sizeof(hash);
     rc = digest_mgr_digest_final(tokdata, sess, FALSE, &context->hash_context,
                                  hash, &hash_len);
@@ -2275,7 +2331,7 @@ CK_RV rsa_hash_pkcs_verify_final(STDLL_TokData_t *tokdata,
     verify_mech.pParameter = NULL;
 
     rc = verify_mgr_init(tokdata, sess, &verify_ctx, &verify_mech, FALSE,
-                         ctx->key);
+                         ctx->key, FALSE);
     if (rc != CKR_OK) {
         TRACE_DEVEL("Verify Mgr Init failed.\n");
         goto done;
@@ -2289,7 +2345,7 @@ done:
         free(octet_str);
     if (ber_data)
         free(ber_data);
-    verify_mgr_cleanup(&verify_ctx);
+    verify_mgr_cleanup(tokdata, sess, &verify_ctx);
 
     return rc;
 }
@@ -2545,7 +2601,7 @@ CK_RV decode_eme_oaep(STDLL_TokData_t *tokdata, CK_BYTE *emData,
         error++;
 
     ps_len = hlen;
-    while ((dbMask[ps_len] == 0x00) && (ps_len < dbMask_len))
+    while ((ps_len < dbMask_len) && (dbMask[ps_len] == 0x00))
         ps_len++;
 
     if ((ps_len >= dbMask_len) ||
@@ -2734,7 +2790,7 @@ CK_RV emsa_pss_verify(STDLL_TokData_t *tokdata,
     /* pkcs1v2.2, Step 10: check DB. */
     i = 0;
     plen = emLen - hlen - pssParms->sLen - 2;
-    while ((buf[i] == 0) && (i < plen))
+    while ((i < plen) && (buf[i] == 0))
         i++;
 
     if ((i != plen) || (buf[i++] != 0x01)) {
@@ -2779,7 +2835,8 @@ CK_RV check_pss_params(CK_MECHANISM *mech, CK_ULONG modlen)
 
     pssParams = (CK_RSA_PKCS_PSS_PARAMS *) mech->pParameter;
 
-    if (mech->ulParameterLen != sizeof(CK_RSA_PKCS_PSS_PARAMS)) {
+    if (mech->ulParameterLen != sizeof(CK_RSA_PKCS_PSS_PARAMS) ||
+        mech->pParameter == NULL) {
         TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
         return CKR_MECHANISM_PARAM_INVALID;
     }
