@@ -21,11 +21,15 @@
 #include "log.h"
 #include "slotmgr.h"
 #include "pkcsslotd.h"
-#include "parser.h"
+#include "cfgparser.h"
+#include "configuration.h"
 
 #define OBJ_DIR "TOK_OBJ"
 #define MD5_HASH_SIZE 16
 
+#define DEF_MANUFID "IBM"
+#define DEF_SLOTDESC    "Linux"
+
 typedef char md5_hash_entry[MD5_HASH_SIZE];
 md5_hash_entry tokname_hash_table[NUMBER_SLOTS_MANAGED];
 
@@ -33,10 +37,11 @@ Slot_Mgr_Shr_t *shmp;           // pointer to the shared memory region.
 int shmid;
 key_t tok;
 Slot_Info_t_64 sinfo[NUMBER_SLOTS_MANAGED];
-Slot_Info_t_64 *psinfo;
 unsigned int NumberSlotsInDB = 0;
+int event_support_disabled = 0;
+
+Slot_Info_t_64 *psinfo;
 
-int socketfd;
 Slot_Mgr_Socket_t socketData;
 
 struct dircheckinfo_s {
@@ -54,17 +59,16 @@ extern BOOL IveDaemonized;
 void DumpSharedMemory(void)
 {
     u_int32 *p;
-    char Buf[PATH_MAX];
+    char buf[4 * 8 + 4];
     u_int32 i;
 
     p = (u_int32 *) shmp;
 
     for (i = 0; i < 15; i++) {
-        sprintf(Buf, "%08X %08X %08X %08X", p[0 + (i * 4)], p[1 + (i * 4)],
+        sprintf(buf, "%08X %08X %08X %08X", p[0 + (i * 4)], p[1 + (i * 4)],
                 p[2 + (i * 4)], p[3 + (i * 4)]);
-        LogLog(Buf);
+        LogLog(buf);
     }
-    return;
 }
 
 int compute_hash(int hash_type, int buf_size, char *buf, char *digest)
@@ -112,7 +116,7 @@ int compute_hash(int hash_type, int buf_size, char *buf, char *digest)
 /** This function does basic sanity checks to make sure the
  *  eco system is in place for opencryptoki to run properly.
  **/
-void run_sanity_checks()
+void run_sanity_checks(void)
 {
     int i, ec, uid = -1;
     struct group *grp = NULL;
@@ -214,6 +218,14 @@ int chk_create_tokdir(Slot_Info_t_64 *psinfo)
     if (!tokdir || strlen(tokdir) == 0)
         return 0;
 
+    /* Check if the path length fits in the max path length
+       (include 2 * / and 0)
+     */
+    if (strlen(CONFIG_PATH) + strlen(tokdir) + strlen(OBJ_DIR) + 3 > PATH_MAX) {
+        fprintf(stderr, "Path name for token object directory too long!\n");
+        return -1;
+    }
+
     proc_umask = umask(0);
 
     /* get 'PKCS11' group id */
@@ -243,6 +255,7 @@ int chk_create_tokdir(Slot_Info_t_64 *psinfo)
            MD5_HASH_SIZE);
 
     /* Create token specific directory */
+    /* sprintf checked above */
     sprintf(tokendir, "%s/%s", CONFIG_PATH, tokdir);
     rc = stat(tokendir, &sbuf);
     if (rc != 0 && errno == ENOENT) {
@@ -268,6 +281,7 @@ int chk_create_tokdir(Slot_Info_t_64 *psinfo)
     }
 
     /* Create TOK_OBJ directory */
+    /* sprintf checked above */
     sprintf(tokendir, "%s/%s/%s", CONFIG_PATH, tokdir, OBJ_DIR);
     rc = stat(tokendir, &sbuf);
     if (rc != 0 && errno == ENOENT) {
@@ -313,6 +327,283 @@ static int create_pid_file(pid_t pid)
     return 0;
 }
 
+static void config_parse_error(int line, int col, const char *msg)
+{
+    ErrLog("Error parsing config file: line %d column %d: %s\n", line, col,
+           msg);
+}
+
+static int do_str(char *slotinfo, size_t size, const char* file,
+                  const char* tok, const char *val, char padding)
+{
+    if (strlen(val) > size) {
+        ErrLog("Error parsing config file '%s': %s has too many characters\n",
+               file, tok);
+        return -1;
+    }
+    memset(slotinfo, padding, size);
+    memcpy(slotinfo, val, strlen(val));
+    return 0;
+}
+
+static int config_parse_slot(const char *config_file,
+                             struct ConfigIdxStructNode *slot)
+{
+    struct ConfigBaseNode *c;
+    int i, slot_no;
+    unsigned int vers;
+    char *str;
+
+    slot_no = slot->idx;
+    DbgLog(DL3, "Slot: %d\n", slot_no);
+
+    if (slot_no >= NUMBER_SLOTS_MANAGED) {
+        ErrLog("Error parsing config file '%s': Slot number %d unsupported! "
+               "Slot number has to be less than %d!", config_file, slot_no,
+               NUMBER_SLOTS_MANAGED);
+        return 1;
+    }
+
+    memset(&sinfo[slot_no].slot_number, 0, sizeof(sinfo[slot_no].slot_number));
+    sinfo[slot_no].slot_number = slot_no;
+
+    confignode_foreach(c, slot->value, i) {
+        DbgLog(DL3, "Config node: '%s' type: %u line: %u\n",
+               c->key, c->type, c->line);
+
+        if (strcmp(c->key, "stdll") == 0 &&
+            (str = confignode_getstr(c)) != NULL) {
+
+            if (do_str((char *)&sinfo[slot_no].dll_location,
+                       sizeof(sinfo[slot_no].dll_location),
+                       config_file, c->key, str, 0))
+                return 1;
+
+            sinfo[slot_no].present = TRUE;
+            continue;
+        }
+
+        if (strcmp(c->key, "description") == 0 &&
+            (str = confignode_getstr(c)) != NULL) {
+            if (do_str((char *)&sinfo[slot_no].pk_slot.slotDescription,
+                       sizeof(sinfo[slot_no].pk_slot.slotDescription),
+                       config_file, c->key, str, ' '))
+                return 1;
+
+            continue;
+        }
+
+        if (strcmp(c->key, "manufacturer") == 0 &&
+            (str = confignode_getstr(c)) != NULL) {
+            if (do_str((char *)&sinfo[slot_no].pk_slot.manufacturerID,
+                       sizeof(sinfo[slot_no].pk_slot.manufacturerID),
+                       config_file, c->key, str, ' '))
+                return 1;
+
+            continue;
+        }
+
+        if (strcmp(c->key, "confname") == 0 &&
+            (str = confignode_getstr(c)) != NULL) {
+            if (do_str((char *)&sinfo[slot_no].confname,
+                       sizeof(sinfo[slot_no].confname),
+                       config_file, c->key, str, 0))
+                return 1;
+
+            continue;
+        }
+
+        if (strcmp(c->key, "tokname") == 0 &&
+            (str = confignode_getstr(c)) != NULL) {
+            if (do_str((char *)&sinfo[slot_no].tokname,
+                       sizeof(sinfo[slot_no].tokname),
+                       config_file, c->key, str, 0))
+                return 1;
+
+            continue;
+        }
+
+        if (strcmp(c->key, "hwversion") == 0 &&
+            confignode_getversion(c, &vers) == 0) {
+            sinfo[slot_no].pk_slot.hardwareVersion.major = vers >> 16;
+            sinfo[slot_no].pk_slot.hardwareVersion.minor = vers & 0xffffu;
+            continue;
+        }
+
+        if (strcmp(c->key, "firmwareversion") == 0 &&
+            confignode_getversion(c, &vers) == 0) {
+            sinfo[slot_no].pk_slot.firmwareVersion.major = vers >> 16;
+            sinfo[slot_no].pk_slot.firmwareVersion.minor = vers & 0xffffu;
+            continue;
+        }
+
+        if (strcmp(c->key, "tokversion") == 0 &&
+            confignode_getversion(c, &sinfo[slot_no].version) == 0)
+            continue;
+
+        ErrLog("Error parsing config file '%s': unexpected token '%s' "
+               "at line %d: \n", config_file, c->key, c->line);
+        return 1;
+    }
+
+    /* set some defaults if user hasn't set these. */
+    if (!sinfo[slot_no].pk_slot.slotDescription[0]) {
+        memset(&sinfo[slot_no].pk_slot.slotDescription[0], ' ',
+               sizeof(sinfo[slot_no].pk_slot.slotDescription));
+        memcpy(&sinfo[slot_no].pk_slot.slotDescription[0],
+               DEF_SLOTDESC, strlen(DEF_SLOTDESC));
+    }
+    if (!sinfo[slot_no].pk_slot.manufacturerID[0]) {
+        memset(&sinfo[slot_no].pk_slot.manufacturerID[0], ' ',
+               sizeof(sinfo[slot_no].pk_slot.manufacturerID));
+        memcpy(&sinfo[slot_no].pk_slot.manufacturerID[0],
+               DEF_MANUFID, strlen(DEF_MANUFID));
+    }
+
+    NumberSlotsInDB++;
+
+    return 0;
+}
+
+static int config_parse_statistics(const char *config_file,
+                                   struct ConfigBareListNode *statistics)
+{
+    struct ConfigBaseNode *c;
+    int i;
+
+    confignode_foreach(c, statistics->value, i) {
+        DbgLog(DL3, "Config node: '%s' type: %u line: %u\n",
+               c->key, c->type, c->line);
+
+        if (c->type == CT_BARE && strcmp(c->key, "off") == 0) {
+            socketData.flags &= ~(FLAG_STATISTICS_ENABLED |
+                                  FLAG_STATISTICS_IMPLICIT |
+                                  FLAG_STATISTICS_INTERNAL);
+            continue;
+        }
+        if (c->type == CT_BARE && strcmp(c->key, "on") == 0) {
+            socketData.flags |= FLAG_STATISTICS_ENABLED;
+            continue;
+        }
+        if (c->type == CT_BARE && strcmp(c->key, "implicit") == 0) {
+            socketData.flags |= FLAG_STATISTICS_IMPLICIT;
+            continue;
+        }
+        if (c->type == CT_BARE && strcmp(c->key, "internal") == 0) {
+            socketData.flags |= FLAG_STATISTICS_INTERNAL;
+            continue;
+        }
+
+        ErrLog("Error parsing config file '%s': unexpected token '%s' "
+               "at line %d: \n", config_file, c->key, c->line);
+        return 1;
+    }
+
+    return 0;
+}
+
+static int config_parse(const char *config_file)
+{
+    FILE *file;
+    struct stat statbuf;
+    struct ConfigBaseNode *c, *config = NULL;
+    struct ConfigIdxStructNode *slot;
+    struct ConfigBareListNode *statistics;
+    int i, ret = 0;
+
+    file = fopen(config_file, "r");
+    if (file == NULL) {
+        ErrLog("Error opening config file '%s': %s\n", config_file,
+               strerror(errno));
+        return -1;
+    }
+
+    if (fstat(fileno(file), &statbuf)) {
+        ErrLog("Error get file information for config file '%s': %s\n",
+               config_file,
+               strerror(errno));
+        fclose(file);
+        return -1;
+    }
+
+    if ((statbuf.st_mode & S_IWOTH)) {
+        ErrLog("Config file %s is world writable, this is not accepted\n",
+               config_file);
+        fclose(file);
+         return -1;
+    }
+
+    ret = parse_configlib_file(file, &config, config_parse_error, 0);
+    fclose(file);
+    if (ret != 0) {
+        ErrLog("Error parsing config file '%s'\n", config_file);
+        goto done;
+    }
+
+    confignode_foreach(c, config, i) {
+        DbgLog(DL3, "Config node: '%s' type: %u line: %u\n",
+               c->key, c->type, c->line);
+
+        if (confignode_hastype(c, CT_FILEVERSION)) {
+            DbgLog(DL0, "Config file version: '%s'\n",
+                   confignode_to_fileversion(c)->base.key);
+            continue;
+        }
+
+        if (confignode_hastype(c, CT_BARECONST)) {
+            if (strcmp(confignode_to_bareconst(c)->base.key,
+                       "disable-event-support") == 0) {
+                event_support_disabled = 1;
+                continue;
+            }
+
+            ErrLog("Error parsing config file '%s': unexpected token '%s' "
+                   "at line %d: \n", config_file, c->key, c->line);
+            ret = -1;
+            break;
+        }
+
+        if (confignode_hastype(c, CT_BARELIST)) {
+            statistics = confignode_to_barelist(c);
+            if (strcmp(statistics->base.key, "statistics") == 0) {
+                ret = config_parse_statistics(config_file, statistics);
+                if (ret != 0)
+                    break;
+                continue;
+            }
+
+            ErrLog("Error parsing config file '%s': unexpected token '%s' "
+                   "at line %d: \n", config_file, c->key, c->line);
+            ret = -1;
+            break;
+        }
+
+        if (confignode_hastype(c, CT_IDX_STRUCT)) {
+            slot = confignode_to_idxstruct(c);
+            if (strcmp(slot->base.key, "slot") == 0) {
+                ret = config_parse_slot(config_file, slot);
+                if (ret != 0)
+                    break;
+                continue;
+            }
+
+            ErrLog("Error parsing config file '%s': unexpected token '%s' "
+                   "at line %d: \n", config_file, c->key, c->line);
+            ret = -1;
+            break;
+        }
+
+        ErrLog("Error parsing config file '%s': unexpected token '%s' "
+               "at line %d: \n", config_file, c->key, c->line);
+        ret = -1;
+        break;
+    }
+
+done:
+    confignode_deepfree(config);
+    return ret;
+}
+
 /*****************************************
  *  main() -
  *      You know what main does.
@@ -349,7 +640,8 @@ int main(int argc, char *argv[], char *envp[])
                GetDebugLevel(), DEBUG_NONE, DEBUG_LEVEL0, DEBUG_LEVEL1);
     }
 
-    ret = load_and_parse(OCK_CONFIG);
+    socketData.flags |= FLAG_STATISTICS_ENABLED;
+    ret = config_parse(OCK_CONFIG);
     if (ret != 0) {
         ErrLog("Failed to read config file.\n");
         return 1;
@@ -397,20 +689,22 @@ int main(int argc, char *argv[], char *envp[])
     if (!XProcUnLock())
         return 4;
 
-    if ((socketfd = CreateListenerSocket()) < 0) {
+    if (!init_socket_server(event_support_disabled)) {
         DestroyMutexes();
         DetachFromSharedMemory();
         DestroySharedMemory();
         return 5;
     }
 
-    if (!InitSocketData(&socketData)) {
-        DetachSocketListener(socketfd);
+    if (!init_socket_data(&socketData)) {
+        term_socket_server();
         DestroyMutexes();
         DetachFromSharedMemory();
         DestroySharedMemory();
         return 6;
     }
+    if (event_support_disabled)
+        socketData.flags |= FLAG_EVENT_SUPPORT_DISABLED;
 
     /* Create customized token directories */
     psinfo = &socketData.slot_info[0];
@@ -426,7 +720,7 @@ int main(int argc, char *argv[], char *envp[])
     if (Daemon) {
         pid_t pid;
         if ((pid = fork()) < 0) {
-            DetachSocketListener(socketfd);
+            term_socket_server();
             DestroyMutexes();
             DetachFromSharedMemory();
             DestroySharedMemory();
@@ -471,7 +765,7 @@ int main(int argc, char *argv[], char *envp[])
      * the daemonization process redefines our handler for (at least) SIGTERM
      */
     if (!SetupSignalHandlers()) {
-        DetachSocketListener(socketfd);
+        term_socket_server();
         DestroyMutexes();
         DetachFromSharedMemory();
         DestroySharedMemory();
@@ -489,10 +783,9 @@ int main(int argc, char *argv[], char *envp[])
      */
 
 #if !defined(NOGARBAGE)
-    printf("Start garbage \n");
     /* start garbage collection thread */
     if (!StartGCThread(shmp)) {
-        DetachSocketListener(socketfd);
+        term_socket_server();
         DestroyMutexes();
         DetachFromSharedMemory();
         DestroySharedMemory();
@@ -512,7 +805,7 @@ int main(int argc, char *argv[], char *envp[])
 #if !(THREADED) && !(NOGARBAGE)
         CheckForGarbage(shmp);
 #endif
-        SocketConnectionHandler(socketfd, 10);
+        socket_connection_handler(10);
     }
 
     /*************************************************************
